import uuid
import marshmallow
from marshmallow import validate, fields


def gen_uuid():
    return str(uuid.uuid4()).replace("-", "")


class BaseSchema(marshmallow.Schema):
    class Meta:
        strict = True
        ordered = True


class FooBodySchema(BaseSchema):
    username = fields.String(required=True, validate=validate.Length(min=1, max=512))
    email = fields.String(required=True, validate=validate.Email())
    some_id = fields.String(missing=gen_uuid, validate=validate.Length(min=1, max=256))
    some_state = fields.Dict(missing=lambda: {})


class BarBodySchema(BaseSchema):
    a_field = fields.String(required=True, validate=validate.Length(min=1, max=512))
    b_field = fields.String(required=True, validate=validate.Length(min=1, max=512))



import random
from wsgiref.simple_server import make_server
from pyramid.config import Configurator
from pyramid.view import view_config, exception_view_config
from pyramid_apispec.helpers import add_pyramid_paths
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin

import validation


@view_config(route_name="users", renderer="json")
def users_post(request):
    """ Some comment.

    ---
    x-extension: value
    post:
      security:
        - APIKeyHeader: []
      tags:
      - "Legacy API"
      summary: "Register user"
      description: ""
      operationId: "users_post"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Foo bar"
        required: true
        schema:
          $ref: "#/definitions/FooBodySchema"
      responses:
          200:
              description: response for 200 code
              schema:
                $ref: "#/definitions/FooBodySchema"
    """
    schema = validation.FooBodySchema()
    struct = schema.loads(request.body or "{}")
    return struct.data


@view_config(route_name="bar_route", request_method="GET", renderer="json")
def bar_get(request):
    """ Return something.

    ---
    get:
        description: some description
        parameters:
          - in: query
            name: offset
            schema:
              type: integer
            description: The number of items to skip
        responses:
            200:
                description: response for 200 code
                schema:
                  $ref: "#/definitions/BarBodySchema"
    """
    return {"a_field": random.random(), "b_field": random.randint(1, 999999)}


@view_config(route_name="bar_route", request_method="POST", renderer="json")
def bar_post(request):
    """ Return something.

    ---
    post:
      tags:
      - "Other API"
      summary: "Send a list of bar's"
      description: ""
      operationId: "bar_post"
      consumes:
      - "application/json"
      produces:
      - "application/json"
      parameters:
      - in: "body"
        name: "body"
        description: "Bar body description"
        required: true
        schema:
          $ref: "#/definitions/BarBodySchema"
      responses:
          200:
              description: response for 200 code
              schema:
                $ref: "#/definitions/BarBodySchema"
    """
    schema = validation.BarBodySchema(many=True)
    struct = schema.loads(request.body or "{}")
    return struct.data


@view_config(route_name="openapi_spec", renderer="json")
def api_spec(request):
    """
    Serve the spec to explorer
    """
    spec = APISpec(title="Some API", version="1.0.0", plugins=[MarshmallowPlugin()])
    # using marshmallow plugin here
    spec.definition("FooBodySchema", schema=validation.FooBodySchema)
    spec.definition("BarBodySchema", schema=validation.BarBodySchema(many=True))

    # inspect the `foo_route` and generate operations from docstring
    add_pyramid_paths(spec, "users", request=request)

    # inspection supports filtering via pyramid add_view predicate arguments
    add_pyramid_paths(spec, "bar_route", request=request)
    my_spec = spec.to_dict()
    # you can further mutate the spec dict to include things like security definitions
    return my_spec


@exception_view_config(context="marshmallow.ValidationError", renderer="json")
def marshmallow_invalid_data(context, request):
    request.response.status = 422
    return context.messages


def build_wsgi_app():
    with Configurator() as config:
        config.add_route("users", "/users")
        config.add_route("bar_route", "/bar")
        config.add_route("openapi_spec", "/openapi.json")

        config.include("pyramid_apispec.views")
        config.pyramid_apispec_add_explorer(spec_route_name="openapi_spec")
        config.scan(".")
        app = config.make_wsgi_app()
    return app


if __name__ == "__main__":
    app = build_wsgi_app()
    server = make_server("0.0.0.0", 6543, app)
    print("visit api explorer at http://127.0.0.1:6543/api-explorer")
    server.serve_forever()



import io
import json

from webtest import TestApp
from app import build_wsgi_app

app = TestApp(build_wsgi_app())

explorer_html = app.get("/api-explorer").text
spec_json = app.get("/openapi.json").json

explorer_html = explorer_html.replace("http://localhost/", "")


with io.open("../gh-pages/index.html", "w") as f:
    f.write(explorer_html)

with io.open("../gh-pages/openapi.json", "w") as f:
    f.write(json.dumps(spec_json, indent=4))



from string import Template
import pkg_resources
import importlib
from pyramid.response import Response
from pyramid_apispec.exceptions import PyramidApiSpecException


def includeme(config):
    settings = config.registry.settings
    config.add_directive(
        "pyramid_apispec_add_explorer", "pyramid_apispec.views:build_api_explorer_view"
    )


def build_api_explorer_view(
    config,
    explorer_route_path="/api-explorer",
    spec_route_name=None,
    script_generator="pyramid_apispec.views:swagger_ui_script_template",
    permission=None,
    **kwargs
):
    """
        Create view that will serve template for swagger UI with proper
        urls substituted

    :param config:
    :param explorer_route_path:
    :param spec_route_name:
    :param script_generator:
    :param permission:
    :param kwargs:
    :return:
    """

    config.add_route("pyramid_apispec.api_explorer_path", explorer_route_path)
    template = pkg_resources.resource_string(
        "pyramid_apispec", "static/index.html"
    ).decode("utf8")
    package, callable = script_generator.split(":")
    imported_package = importlib.import_module(package)

    def swagger_ui_template_view(request):
        if not spec_route_name:
            raise PyramidApiSpecException(
                "spec_route_name argument needs to be present"
            )

        script_callable = getattr(imported_package, callable)
        version = kwargs.get("ui_version", "3.17.1")
        html = Template(template).safe_substitute(
            swagger_ui_script=script_callable(request, spec_route_name, **kwargs),
            version=version,
        )
        return Response(html)

    config.add_view(
        swagger_ui_template_view,
        permission=permission,
        route_name="pyramid_apispec.api_explorer_path",
    )


def swagger_ui_script_template(request, spec_route_name, **kwargs):
    """
    Generates the <script> code that bootstraps Swagger UI, it will be injected
    into index template

    :param request:
    :param spec_route_name:
    :return:
    """
    template = pkg_resources.resource_string(
        "pyramid_apispec", "static/index_script_template.html"
    ).decode("utf8")
    return Template(template).safe_substitute(
        swagger_spec_url=request.route_url(spec_route_name)
    )



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
from io import open

from pyramid_apispec import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

version = '{}.{}.{}'.format(__version__['major'],
                            __version__['minor'],
                            __version__['patch'])

setup(
    name='pyramid_apispec',
    version=version,
    description='Pyramid plugin for openapi spec generation ',
    long_description=long_description,
    license='BSD',
    long_description_content_type='text/markdown',
    url='https://github.com/ergo/pyramid_apispec',
    author='Marcin Lulek',
    author_email='marcin@webreactor.eu',
    classifiers=[  # Optional
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Framework :: Pyramid',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='pyramid apispec marshmallow rest restful',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    package_data={
        'pyramid_apispec': ['static/*.*'],
        '': ['LICENSE'],
    },
    install_requires=['apispec[yaml]>=1.0.0b1'],
    setup_requires=["pytest-runner"],
    extras_require={
        'dev': ['coverage', 'pytest', 'pyramid', 'tox', 'webtest'],
        'demo': ['marshmallow==2.15.3', 'pyramid', 'apispec', 'webtest'],
        'lint': ['black'],
    }
)



# -*- coding: utf-8 -*-
"""Pyramid helper. Includes a path helper that allows you to pass a route name
for inspection. Inspects URL rules and view docstrings.

Inspecting a route and its view::

    from pyramid_apispec.helpers import add_pyramid_paths

    @view_config(route_name='foo_route', renderer='json')
    def foo_view():
        \"""A greeting endpoint.

        ---
        x-extension: value
        get:
            description: get a greeting
            responses:
                200:
                    description: a pet to be returned
                    schema:
                        $ref: #/definitions/SomeFooBody
        \"""
        return 'hi'

    @view_config(route_name='openapi_spec', renderer='json')
    def api_spec(request):
        spec = APISpec(
            title='Some API',
            version='1.0.0',
            plugins=[
                'apispec.ext.marshmallow'
            ],
        )
        # using marshmallow plugin here
        spec.definition('SomeFooBody', schema=MarshmallowSomeFooBodySchema)

        # inspect the `foo_route` and generate operations from docstring
        add_pyramid_paths(spec, 'foo_route', request=request)

        # inspection supports filtering via pyramid add_view predicate arguments
        add_pyramid_paths(
            spec, 'bar_route', request=request, request_method='post')
        return spec.to_dict()


"""
from __future__ import absolute_import

from apispec.yaml_utils import load_operations_from_docstring, load_yaml_from_docstring

from pyramid.threadlocal import get_current_request

# py 2/3 compat
try:
    import basestring

    string_type = basestring
except ImportError:
    string_type = str


def is_string(val):
    return isinstance(val, string_type)


ALL_METHODS = ("get", "post", "put", "patch", "delete", "head", "options")


def clean_part(part):
    # cleanup regex from route params
    if part.startswith("{") and part.endswith("}") and ":" in part:
        part = part.split(":")[0] + "}"
    return part


def reformat_pattern(pattern):
    if not pattern.startswith("/"):
        pattern = "/%s" % pattern
    parts = pattern.split("/")
    return "/".join([clean_part(p) for p in parts])


def add_pyramid_paths(
    spec,
    route_name,
    request=None,
    request_method=None,
    operations=None,
    autodoc=True,
    **kwargs
):
    """

    Adds a route and view info to spec

    :param spec:
        ApiSpec object
    :param route_name:
        Route name to inspect
    :param request:
        Request object, if `None` then `get_current_request()` will be used
    :param request_method:
        Request method predicate
    :param operations:
        Operations dict that will be used instead of introspection
    :param autodoc:
        Include information about endpoints without markdown docstring
    :param kwargs:
        Additional kwargs for predicate matching
    :return:

    """
    if request is None:
        request = get_current_request()

    registry = request.registry
    introspector = registry.introspector
    route = introspector.get("routes", route_name)
    introspectables = introspector.related(route)

    # needs to be rewritten to internal name
    if request_method:
        kwargs["request_methods"] = request_method

    for maybe_view in introspectables:
        if not is_view(maybe_view) or not check_methods_matching(maybe_view, **kwargs):
            continue

        pattern = route["pattern"]
        pattern = reformat_pattern(pattern)
        spec.add_path(
            pattern, operations=get_operations(maybe_view, operations, autodoc=autodoc)
        )


def is_view(introspectable):
    return introspectable.category_name == "views"


def check_methods_matching(view, **kwargs):
    for kw in kwargs.keys():
        # request_methods can be either a list of strings or a string
        # so lets normalize via sets
        if kw == "request_methods":
            if is_string(kwargs[kw]):
                kwargs[kw] = [kwargs[kw]]
            methods = view.get(kw) or ALL_METHODS
            if is_string(methods):
                methods = [methods]
            if not set(kwargs[kw] or []).intersection(methods):
                return False
        else:
            if not view.get(kw) == kwargs[kw]:
                return False
    return True


def get_operations(view, operations, autodoc=True):
    if operations is not None:
        return operations

    operations = {}

    # views can be class based
    if view.get("attr"):
        global_meta = load_operations_from_docstring(view["callable"].__doc__)
        if global_meta:
            operations.update(global_meta)
        f_view = getattr(view["callable"], view["attr"])
    # or just function callables
    else:
        f_view = view.get("callable")

    methods = view.get("request_methods")
    view_operations = load_operations_from_docstring(f_view.__doc__)
    if not view_operations:
        view_operations = {}
        if is_string(methods):
            methods = [methods]
        if not methods:
            methods = ALL_METHODS[:]
        operation = load_yaml_from_docstring(f_view.__doc__)
        if operation:
            for method in methods:
                view_operations[method.lower()] = operation
        elif autodoc:
            for method in methods:
                view_operations.setdefault(method.lower(), {"responses": {}})
    operations.update(view_operations)

    return operations
